@@ -24,6 +24,16 @@
 
 #include <sys/param.h>
 
+#ifdef __arch_v32
+
+/* FIXME: Find all uses and intended uses and specify delay-slot there
+   instead.  */
+#define CALL_FN(x)				\
+	"bsr " #x "\n\t"			\
+	"nop"
+
+#else /* not __arch_v32 */
+
 #ifdef __PIC__
 /* Due to limitations in the binutils CRIS port, this only works if x is
    a local symbol in the same section.  */
@@ -35,12 +45,26 @@
 # define CALL_FN(x) "jsr " #x
 #endif
 
+#endif /* not __arch_v32 */
+
 /* Return nonzero iff ELF header is compatible with the running host.  */
 
 static inline int
 elf_machine_matches_host (const Elf32_Ehdr *ehdr)
 {
-  return ehdr->e_machine == EM_CRIS;
+  return ehdr->e_machine == EM_CRIS
+#ifdef __arch_v32
+    && (((ehdr->e_flags & EF_CRIS_VARIANT_MASK)
+	 == EF_CRIS_VARIANT_V32)
+	|| (ehdr->e_flags & EF_CRIS_VARIANT_MASK)
+	 == EF_CRIS_VARIANT_COMMON_V10_V32)
+#else
+    && (((ehdr->e_flags & EF_CRIS_VARIANT_MASK)
+	 == EF_CRIS_VARIANT_ANY_V0_V10)
+	|| (ehdr->e_flags & EF_CRIS_VARIANT_MASK)
+	== EF_CRIS_VARIANT_COMMON_V10_V32)
+#endif
+    ;
 }
 
 /* Return the link-time address of _DYNAMIC.  Conveniently, this is the
@@ -73,8 +97,31 @@ static inline Elf32_Addr
 elf_machine_load_address (void)
 {
   Elf32_Addr gotaddr_diff;
+#ifdef __arch_v32
+  extern char ___CRISv32_dummy[] __asm__ ("_dl_start");
+  __asm__ ("addo.w _dl_start:GOT16,$r0,$acr\n\t"
+	   "lapc _dl_start,%0\n\t"
+	   "sub.d [$acr],%0"
+	   /* For v32, we need to force GCC to have R0 loaded with
+	      _GLOBAL_OFFSET_TABLE_ at this point, which might not
+	      otherwise have happened in the caller.  (For v10, it's
+	      loaded for non-global variables too, so we don't need
+	      anything special there.)  We accomplish this by faking the
+	      address of a global variable (as seen by GCC) as input to
+	      the asm; that address calculation goes through the GOT.
+	      Use of this function happens before we've filled in the
+	      GOT, so the address itself will not be correctly
+	      calculated, therefore we don't use any symbol whose
+	      address may be re-used later on.  Let's just reuse the
+	      _dl_start symbol, faking it as a global by renaming it as
+	      another variable through an asm.  */
+	   : "=r" (gotaddr_diff)
+	   : "g" (___CRISv32_dummy)
+	   : "acr");
+#else
   __asm__ ("sub.d [$r0+_dl_start:GOT16],$r0,%0\n\t"
 	   "add.d _dl_start:GOTOFF,%0" : "=r" (gotaddr_diff));
+#endif
   return gotaddr_diff;
 }
 
@@ -130,29 +177,64 @@ elf_machine_runtime_setup (struct link_map *l, int lazy, int profile)
    We get here with the offset into the relocation table pushed on stack,
    and the link map in MOF.  */
 
+#ifdef __arch_v32
 #define TRAMPOLINE_TEMPLATE(tramp_name, fixup_name) \
-"; Trampoline for " #fixup_name "
-	.globl " #tramp_name "
-	.type " #tramp_name ", @function
-" #tramp_name ":
-	push	$r13
-	push	$r12
-	push	$r11
-	push	$r10
-	push	$r9
-	push	$srp
-	move.d	[$sp+6*4],$r11
-	move	$mof,$r10
-	" CALL_FN (fixup_name) "
-	move.d	$r10,[$sp+6*4]
-	pop	$srp
-	pop	$r9
-	pop	$r10
-	pop	$r11
-	pop	$r12
-	pop	$r13
-	jump	[$sp+]
+"; Trampoline for " #fixup_name	"					\n\
+	.globl " #tramp_name "						\n\
+	.type " #tramp_name ", @function				\n\
+" #tramp_name ":							\n\
+	subq	4,$sp							\n\
+	move.d	$r13,[$sp]						\n\
+	subq	4,$sp							\n\
+	move.d	$r12,[$sp]						\n\
+	subq	4,$sp							\n\
+	move.d	$r11,[$sp]						\n\
+	subq	4,$sp							\n\
+	addoq	4*4,$sp,$acr						\n\
+	move.d	$r10,[$sp]						\n\
+	subq	4,$sp							\n\
+	move	$mof,$r10						\n\
+	move.d	$r9,[$sp]						\n\
+	subq	4,$sp							\n\
+	move.d	[$acr],$r11						\n\
+	move	$srp,[$sp]						\n\
+	" CALL_FN (fixup_name) "					\n\
+	move.d	$r10,$acr						\n\
+	move	[$sp+],$srp						\n\
+	move.d	[$sp+],$r9						\n\
+	move.d	[$sp+],$r10						\n\
+	move.d	[$sp+],$r11						\n\
+	move.d	[$sp+],$r12						\n\
+	move.d	[$sp+],$r13						\n\
+	jump	$acr							\n\
+	addq	4,$sp							\n\
 	.size " #tramp_name ", . - " #tramp_name "\n"
+#else
+#define TRAMPOLINE_TEMPLATE(tramp_name, fixup_name) \
+"; Trampoline for " #fixup_name "					\n\
+	.globl " #tramp_name "						\n\
+	.type " #tramp_name ", @function				\n\
+" #tramp_name ":							\n\
+	push	$r13							\n\
+	push	$r12							\n\
+	push	$r11							\n\
+	push	$r10							\n\
+	push	$r9							\n\
+	push	$srp							\n\
+	move.d	[$sp+6*4],$r11						\n\
+	move	$mof,$r10						\n\
+	" CALL_FN (fixup_name) "					\n\
+	move.d	$r10,[$sp+6*4]						\n\
+	pop	$srp							\n\
+	pop	$r9							\n\
+	pop	$r10							\n\
+	pop	$r11							\n\
+	pop	$r12							\n\
+	pop	$r13							\n\
+	jump	[$sp+]							\n\
+	.size " #tramp_name ", . - " #tramp_name "\n"
+#endif /* not __arch_v32 */
+
 #ifndef PROF
 #define ELF_MACHINE_RUNTIME_TRAMPOLINE \
 asm (TRAMPOLINE_TEMPLATE (_dl_runtime_resolve, fixup) \
@@ -173,62 +255,126 @@ asm (TRAMPOLINE_TEMPLATE (_dl_runtime_resolve, fixup) \
    The C function `_dl_start' is the real entry point;
    its return value is the user program's entry point.  */
 
+#ifdef __arch_v32
 #define RTLD_START asm ("\
-	.text
-	.globl	_start
-	.type	_start,@function
-_start:
-	move.d	$sp,$r10
-	" CALL_FN (_dl_start) "
-	/* FALLTHRU */
-
-	.globl _dl_start_user
-	.type _dl_start_user,@function
-_dl_start_user:
-	; Save the user entry point address in R1.
-	move.d	$r10,$r1
-	; Point R0 at the GOT.
-	move.d	$pc,$r0
-	sub.d	.:GOTOFF,$r0
-	; Remember the highest stack address.
-	move.d	[$r0+__libc_stack_end:GOT16],$r13
-	move.d	$sp,[$r13]
-	; See if we were run as a command with the executable file
-	; name as an extra leading argument.
-	move.d	[$r0+_dl_skip_args:GOT16],$r13
-	move.d	[$r13],$r9
-	; Get the original argument count
-	move.d	[$sp],$r11
-	; Subtract _dl_skip_args from it.
-	sub.d	$r9,$r11
-	; Adjust the stack pointer to skip _dl_skip_args words.
-	addi	$r9.d,$sp
-	; Put the new argc in place as expected by the user entry.
-	move.d	$r11,[$sp]
-	; Call _dl_init (struct link_map *main_map, int argc, char **argv, char **env)
-	;  env: skip scaled argc and skip stored argc and NULL at end of argv[].
-	move.d	$sp,$r13
-	addi	$r11.d,$r13
-	addq	8,$r13
-	;  argv: skip stored argc.
-	move.d	$sp,$r12
-	addq	4,$r12
-	;  main_map: at _dl_loaded.
-	move.d	[$r0+_dl_loaded:GOT16],$r9
-	move.d	[$r9],$r10
-	move.d	_dl_init:PLTG,$r9
-	add.d	$r0,$r9
-	jsr	$r9
-	; Pass our finalizer function to the user in R10.
-	move.d [$r0+_dl_fini:GOT16],$r10
-	; Terminate the frame-pointer.
-	moveq	0,$r8
-	; Cause SEGV if user entry returns.
-	move	$r8,$srp
-	; Jump to the user's entry point.
-	jump	$r1
-	.size _dl_start_user, . - _dl_start_user
+	.text								\n\
+	.globl	_start							\n\
+	.type	_start,@function					\n\
+_start:									\n\
+	move.d	$sp,$r10						\n\
+	" CALL_FN (_dl_start) "						\n\
+	/* FALLTHRU */							\n\
+									\n\
+	.globl _dl_start_user						\n\
+	.type _dl_start_user,@function					\n\
+_dl_start_user:								\n\
+	; Save the user entry point address in R1.			\n\
+	move.d	$r10,$r1						\n\
+	; Point R0 at the GOT.						\n\
+	lapc	_GLOBAL_OFFSET_TABLE_,$r0				\n\
+	; Remember the highest stack address.				\n\
+	addo.w	__libc_stack_end:GOT16,$r0,$acr				\n\
+	move.d	[$acr],$r13						\n\
+	move.d	$sp,[$r13]						\n\
+	; See if we were run as a command with the executable file	\n\
+	; name as an extra leading argument.				\n\
+	addo.w	_dl_skip_args:GOT16,$r0,$acr				\n\
+	move.d	[$acr],$r13						\n\
+	move.d	[$r13],$r9						\n\
+	; Get the original argument count				\n\
+	move.d	[$sp],$r11						\n\
+	; Subtract _dl_skip_args from it.				\n\
+	sub.d	$r9,$r11						\n\
+	; Adjust the stack pointer to skip _dl_skip_args words.		\n\
+	addi	$r9.d,$sp						\n\
+	; Put the new argc in place as expected by the user entry.	\n\
+	move.d	$r11,[$sp]						\n\
+	; Call _dl_init (struct link_map *main_map, int argc,		\n\
+	;		 char **argv, char **env)			\n\
+	;  env: skip scaled argc and skip stored argc and NULL		\n\
+	;  at end of argv[].						\n\
+	move.d	$sp,$r13						\n\
+	addi	$r11.d,$r13						\n\
+	addq	8,$r13							\n\
+	;  argv: skip stored argc.					\n\
+	move.d	$sp,$r12						\n\
+	addq	4,$r12							\n\
+	;  main_map: at _dl_loaded.					\n\
+	addo.w	_dl_loaded:GOT16,$r0,$acr				\n\
+	move.d	[$acr],$r9						\n\
+	bsr	_dl_init:PLT						\n\
+	move.d	[$r9],$r10						\n\
+	; Pass our finalizer function to the user in R10.		\n\
+	addo.w	_dl_fini:GOT16,$r0,$acr					\n\
+	move.d [$acr],$r10						\n\
+	; Terminate the frame-pointer.					\n\
+	moveq	0,$r8							\n\
+	; Jump to the user's entry point.				\n\
+	jump	$r1							\n\
+	; Cause SEGV if user entry returns.				\n\
+	move	$r8,$srp						\n\
+	.size _dl_start_user, . - _dl_start_user			\n\
 	.previous");
+#else  /* not __arch_v32 */
+#define RTLD_START asm ("\
+	.text								\n\
+	.globl	_start							\n\
+	.type	_start,@function					\n\
+_start:									\n\
+	move.d	$sp,$r10						\n\
+	" CALL_FN (_dl_start) "						\n\
+	/* FALLTHRU */							\n\
+									\n\
+	.globl _dl_start_user						\n\
+	.type _dl_start_user,@function					\n\
+_dl_start_user:								\n\
+	; Save the user entry point address in R1.			\n\
+	move.d	$r10,$r1						\n\
+	; Point R0 at the GOT.						\n\
+	move.d	$pc,$r0							\n\
+	sub.d	.:GOTOFF,$r0						\n\
+	; Remember the highest stack address.				\n\
+	move.d	[$r0+__libc_stack_end:GOT16],$r13			\n\
+	move.d	$sp,[$r13]						\n\
+	; See if we were run as a command with the executable file	\n\
+	; name as an extra leading argument.				\n\
+	move.d	[$r0+_dl_skip_args:GOT16],$r13				\n\
+	move.d	[$r13],$r9						\n\
+	; Get the original argument count				\n\
+	move.d	[$sp],$r11						\n\
+	; Subtract _dl_skip_args from it.				\n\
+	sub.d	$r9,$r11						\n\
+	; Adjust the stack pointer to skip _dl_skip_args words.		\n\
+	addi	$r9.d,$sp						\n\
+	; Put the new argc in place as expected by the user entry.	\n\
+	move.d	$r11,[$sp]						\n\
+	; Call _dl_init (struct link_map *main_map, int argc,		\n\
+	;		 char **argv, char **env)			\n\
+	;  env: skip scaled argc and skip stored argc and NULL		\n\
+	;  at end of argv[].						\n\
+	move.d	$sp,$r13						\n\
+	addi	$r11.d,$r13						\n\
+	addq	8,$r13							\n\
+	;  argv: skip stored argc.					\n\
+	move.d	$sp,$r12						\n\
+	addq	4,$r12							\n\
+	;  main_map: at _dl_loaded.					\n\
+	move.d	[$r0+_dl_loaded:GOT16],$r9				\n\
+	move.d	[$r9],$r10						\n\
+	move.d	_dl_init:PLTG,$r9					\n\
+	add.d	$r0,$r9							\n\
+	jsr	$r9							\n\
+	; Pass our finalizer function to the user in R10.		\n\
+	move.d [$r0+_dl_fini:GOT16],$r10				\n\
+	; Terminate the frame-pointer.					\n\
+	moveq	0,$r8							\n\
+	; Cause SEGV if user entry returns.				\n\
+	move	$r8,$srp						\n\
+	; Jump to the user's entry point.				\n\
+	jump	$r1							\n\
+	.size _dl_start_user, . - _dl_start_user			\n\
+	.previous");
+#endif /* not __arch_v32 */
 
 /* Nonzero if TYPE describes a relocation that should
    skip the executable when looking up the symbol value.  */
